from lib2to3.pytree import Base
from pyexpat.errors import XML_ERROR_INVALID_TOKEN
import torch.nn as nn
from astrape.utilities.utils import *
from astrape.exceptions.exceptions import *
from astrape.models.model_buildingblocks import *
from astrape.utilities.dataloader_lightning import *
import pytorch_lightning as pl
from sklearn.model_selection import  train_test_split
import numpy as np
from typing import Union, List, Dict, Any, Optional, cast, overload
from astrape.utilities.utils_lightning import *
from sklearn.base import BaseEstimator
from astrape.model_selection import CrossValidation
import inspect
from sklearn.model_selection import RandomizedSearchCV, GridSearchCV
import scipy
from torch.utils.tensorboard import SummaryWriter
from datetime import datetime
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
import warnings
warnings.filterwarnings(action='ignore')
from pytorch_lightning.utilities.seed import seed_everything as pl_set_seed

applicable_val_metric_list = APPLICABLE_VAL_METRIC_LIST
applicable_optimizer_type_list = APPLICABLE_OPTIMIZER_TYPE_LIST




class Experiment(BaseExperiment):
    def __init__(
        self,
        project_name : str,
        X : "np.ndarray",
        y : "np.ndarray",
        n_classes : Optional[int] = None,
        X_test : Optional[np.ndarray] = None,
        y_test : Optional[np.ndarray] = None,
        test_size : float = 0.01,
        stack_models : bool = True,
        random_state : int = 0,
        path : Optional[str] = None
    ) -> None:
       
        super().__init__(
            project_name=project_name,
            X=X,
            y=y,
            n_classes=n_classes,
            X_test=X_test,
            y_test=y_test,
            test_size=test_size,
            stack_models=stack_models,
            random_state=random_state,
            path=path
        )
    #######################################################################  cross_validation (Begins)  ####################################################################### 
    def cross_validation(
        self,
        model_type : Union["BaseEstimator", "pl.LightningModule"],
        trainer_config : Optional[Dict],
        parameters : Dict[str, List[Any]],
        search_type : str = "random",
        val_metric : str = 'val/acc',
        n_random_search : Optional[int] = None 
    ):
        pl_set_seed(self.random_state)
        np.random.seed(self.random_state)
        if pl.LightningModule in inspect.getmro(model_type) or nn.Module in inspect.getmro(model_type):
            
            best_checkpoint, info = CrossValidation(
                random_state=self.random_state,
                dims=self.dims,
                n_classes=self.n_classes,
                data_module=self.data_module,
                trainer_config=trainer_config,
                log_dir=self.exp_path+"/logs/",
                ckpt_dir=self.exp_path+"/checkpoints/",
                model_type=model_type,
                parameters=parameters,
                val_metric=val_metric,
                search_type=search_type,
                n_random_search=n_random_search
            )


            self.model = model_type.load_from_checkpoint(best_checkpoint)
            
            self.model_metadata = self.set_model_metadata(self.model, info['best_hparams'])
            best_hparams = info['best_hparams']
            
        elif BaseEstimator in inspect.getmro(model_type):
            param_grid = {}
            for hparam, inner_dict in parameters.items():
                for key, value in inner_dict.items():
                    if key == "value":
                        param_grid.update({hparam : [value]})
                    if key == "values":
                        param_grid.update({hparam : value})
                    elif "distribution" in inner_dict.keys():
                        dist_name = inner_dict["distribution"]
                        if dist_name == "uniform":
                            values_for_param_grid = np.random.uniform(low=inner_dict["min"], high=inner_dict["max"], size=4)
                        elif dist_name == "log_uniform":
                            values_for_param_grid = scipy.stats.reciprocal(np.exp(inner_dict["min"]), np.exp(inner_dict["max"])).rvs(4)
                        elif dist_name == "normal":
                            values_for_param_grid = np.random.normal(loc=inner_dict["mu"], scale=inner_dict["sigma"], size=4)
                        elif dist_name == "log_normal":
                            values_for_param_grid = np.random.normal(mean=inner_dict["mu"], sigma=inner_dict["sigma"], size=4)
                        else:
                            raise NotImplementedError("The type of distribution in parameters is not supported in expytorch_lightning v.0.0.0.")
                    param_grid.update({hparam : values_for_param_grid})

                if search_type != "random" or search_type != "grid":
                    search_type = "random"
                
                if search_type == "random":
                    
                    clf = RandomizedSearchCV(model_type, param_grid, random_state=self.random_state)
                    
                    search = clf.fit(self.X, self.y)
                    best_hparams = search.best_params_
                    best_checkpoint = search.best_estimator_

                    self.model = self.set_model(**info['best_hparams'])
                    self.model_metadata = self.set_model_metadata(info['best_hparams'])

                elif search_type == "grid":
                    for key in inner_dict.keys():
                        if key == "distribution":
                            raise ValueError("v 0.0.0 does not support 'distribution' in GridSearchCV.")

                    clf = GridSearchCV(model_type, param_grid, random_state=self.random_state)
                    
                    search = clf.fit(self.X, self.y)
                    best_hparams = search.best_params_
                    best_checkpoint = search.best_estimator_

                    self.model = self.set_model(**info['best_hparams'])
                    self.model_metadata = self.set_model_metadata(info['best_hparams'])

            return best_checkpoint, best_hparams         
        
    